from .base import ObjectListModel, BaseModel

class JournalList(ObjectListModel):

    def __init__(self):
        super().__init__(list=[], listObject=Journal)

class Journal(BaseModel):

    def __init__(self,
        ID=None,
        AllowVariableCurrency=None,
        AllowVariableExchangeRate=None,
        AllowVAT=None,
        AutoSave=None,
        Bank=None,
        BankAccountBICCode=None,
        BankAccountCountry=None,
        BankAccountDescription=None,
        BankAccountIBAN=None,
        BankAccountID=None,
        BankAccountIncludingMask=None,
        BankAccountUseSEPA=None,
        BankAccountUseSepaDirectDebit=None,
        BankName=None,
        Code=None,
        Created=None,
        Creator=None,
        CreatorFullName=None,
        Currency=None,
        CurrencyDescription=None,
        Description=None,
        Division=None,
        GLAccount=None,
        GLAccountCode=None,
        GLAccountDescription=None,
        GLAccountType=None,
        IsBlocked=None,
        Modified=None,
        Modifier=None,
        ModifierFullName=None,
        PaymentInTransitAccount=None,
        PaymentServiceAccountIdentifier=None,
        PaymentServiceProvider=None,
        PaymentServiceProviderName=None,
        Type=None
    ):

        super().__init__()

        self.ID = ID
        self.AllowVariableCurrency = AllowVariableCurrency
        self.AllowVariableExchangeRate = AllowVariableExchangeRate
        self.AllowVAT = AllowVAT
        self.AutoSave = AutoSave
        self.Bank = Bank
        self.BankAccountBICCode = BankAccountBICCode
        self.BankAccountCountry = BankAccountCountry
        self.BankAccountDescription = BankAccountDescription
        self.BankAccountIBAN = BankAccountIBAN
        self.BankAccountID = BankAccountID
        self.BankAccountIncludingMask = BankAccountIncludingMask
        self.BankAccountUseSEPA = BankAccountUseSEPA
        self.BankAccountUseSepaDirectDebit = BankAccountUseSepaDirectDebit
        self.BankName = BankName
        self.Code = Code
        self.Created = Created
        self.Creator = Creator
        self.CreatorFullName = CreatorFullName
        self.Currency = Currency
        self.CurrencyDescription = CurrencyDescription
        self.Description = Description
        self.Division = Division
        self.GLAccount = GLAccount
        self.GLAccountCode = GLAccountCode
        self.GLAccountDescription = GLAccountDescription
        self.GLAccountType = GLAccountType
        self.IsBlocked = IsBlocked
        self.Modified = Modified
        self.Modifier = Modifier
        self.ModifierFullName = ModifierFullName
        self.PaymentInTransitAccount = PaymentInTransitAccount
        self.PaymentServiceAccountIdentifier = PaymentServiceAccountIdentifier
        self.PaymentServiceProvider = PaymentServiceProvider
        self.PaymentServiceProviderName = PaymentServiceProviderName
        self.Type = Type


from .base import APIEndpoint

from exactonline.models.documents import Document,DocumentAttachment, DocumentAttachmentList, DocumentList, DocumentType, DocumentTypeCategory, DocumentTypeCategoryList, DocumentTypeList
from exactonline.utils import encodeFileToB64, getFileName

class DocumentMethods(APIEndpoint):

    def __init__(self, api):
        super().__init__(api, 'documents', Document, DocumentList)
    
    def list(self, select=[]):
        url = "{endpoint}/Documents".format(endpoint=self.endpoint)
        if select: url = '{url}/Documents&$select={select}'.format(url=url, select=",".join(select))

        status, headers, respJson = self.api.get(url)

        if status != 200: return DocumentList().parseError(respJson)

        return DocumentList().parse(respJson['d']['results'])
    
    def listTypes(self, select=[]):
        url = "{endpoint}/DocumentTypes".format(endpoint=self.endpoint)
        if select: url = '{url}/DocumentTypes&$select={select}'.format(url=url, select=",".join(select))

        status, headers, respJson = self.api.get(url)

        if status != 200: return DocumentTypeList().parseError(respJson)
        
        return DocumentTypeList().parse(respJson['d']['results'])

    def listTypeCategories(self, select=[]):
        url = "{endpoint}/DocumentTypeCategories".format(endpoint=self.endpoint)
        if select: url = '{url}/DocumentTypeCategories&$select={select}'.format(url=url, select=",".join(select))

        status, headers, respJson = self.api.get(url)

        if status != 200: return DocumentTypeCategoryList().parseError(respJson)
        
        return DocumentTypeCategoryList().parse(respJson['d']['results'])

    def get(self, id, select=[]):

        url = "{endpoint}/Documents?$filter=ID eq guid'{id}'".format(endpoint=self.endpoint, id=id)
        if select: url = '{url}&$select={select}'.format(url=url, select=",".join(select))

        status, headers, respJson = self.api.get(url)

        if status != 200: return Document().parseError(respJson)

        return Document().parse(respJson['d']['results'][0])

    def create(self, document, attachments=[]):
    
        url = "{endpoint}/Documents".format(endpoint=self.endpoint)
        data = document.getJSON()

        # 1: create Document
        status, headers, respJson = self.api.post(url, data)
        if status not in [200, 201]: return Document().parseError(respJson)
        
        document = Document().parse(respJson['d'])

        # 2: create and link Attachments
        for attachment in attachments:
            dAttachment = DocumentAttachment(
                Attachment=encodeFileToB64(attachment),
                Document=document.ID,
                FileName=getFileName(attachment)
            )

            data = dAttachment.getJSON()
            url = "{endpoint}/DocumentAttachments".format(endpoint=self.endpoint)
            status, headers, respJson = self.api.post(url, data)
            
            if status in [200, 201]: document.Attachments.add(dAttachment)
        
        return document


from .base import APIEndpoint

from exactonline.models.contacts import Contact, ContactList

class ContactMethods(APIEndpoint):

    def __init__(self, api):
        super().__init__(api, 'crm/Contacts', Contact, ContactList)


from .base import APIEndpoint

from exactonline.models.accounts import Account, AccountList

class AccountMethods(APIEndpoint):

    def __init__(self, api):
        super().__init__(api, 'crm/Accounts', Account, AccountList)


import base64
import os

def getFileName(path):
    return os.path.basename(path)

def encodeFileToB64(path):

    f = open(path, 'rb')
    fRead = f.read()
    fEncode = base64.encodebytes(fRead)
    fString = fEncode.decode('utf-8')

    return fString


from .base import ObjectListModel, BaseModel

class SalesEntryList(ObjectListModel):

    def __init__(self):
        super().__init__(list=[], listObject=SalesEntry)
    

class SalesEntry(BaseModel):

    def __init__(self,
        EntryID=None,
        AmountDC=None,
        AmountFC=None,
        BatchNumber=None,
        Created=None,
        Creator=None,
        CreatorFullName=None,
        Currency=None,
        Customer=None,
        CustomerName=None,
        Description=None,
        Division=None,
        Document=None,
        DocumentNumber=None,
        DocumentSubject=None,
        DueDate=None,
        EntryDate=None,
        EntryNumber=None,
        ExternalLinkDescription=None,
        ExternalLinkReference=None,
        GAccountAmountFC=None,
        InvoiceNumber=None,
        IsExtraDuty=None,
        Journal=None,
        JournalDescription=None,
        Modified=None,
        Modifier=None,
        ModifierFullName=None,
        OrderNumber=None,
        PaymentCondition=None,
        PaymentConditionDescription=None,
        PaymentReference=None,
        ProcessNumber=None,
        Rate=None,
        ReportingPeriod=None,
        ReportingYear=None,
        Reversal=None,
        SalesEntryLines=None,
        Status=None,
        StatusDescription=None,
        Type=None,
        TypeDescription=None,
        VATAmountDC=None,
        VATAmountFC=None,
        WithholdingTaxAmountDC=None,
        WithholdingTaxBaseAmount=None,
        WithholdingTaxPercentage=None,
        YourRef=None
    ):

        super().__init__()

        self.EntryID = EntryID
        self.AmountDC = AmountDC
        self.AmountFC = AmountFC
        self.BatchNumber = BatchNumber
        self.Created = Created
        self.Creator = Creator
        self.CreatorFullName = CreatorFullName
        self.Currency = Currency
        self.Customer = Customer
        self.CustomerName = CustomerName
        self.Description = Description
        self.Division = Division
        self.Document = Document
        self.DocumentNumber = DocumentNumber
        self.DocumentSubject = DocumentSubject
        self.DueDate = DueDate
        self.EntryDate = EntryDate
        self.EntryNumber = EntryNumber
        self.ExternalLinkDescription = ExternalLinkDescription
        self.ExternalLinkReference = ExternalLinkReference
        self.GAccountAmountFC = GAccountAmountFC
        self.InvoiceNumber = InvoiceNumber
        self.IsExtraDuty = IsExtraDuty
        self.Journal = Journal
        self.JournalDescription = JournalDescription
        self.Modified = Modified
        self.Modifier = Modifier
        self.ModifierFullName = ModifierFullName
        self.OrderNumber = OrderNumber
        self.PaymentCondition = PaymentCondition
        self.PaymentConditionDescription = PaymentConditionDescription
        self.PaymentReference = PaymentReference
        self.ProcessNumber = ProcessNumber
        self.Rate = Rate
        self.ReportingPeriod = ReportingPeriod
        self.ReportingYear = ReportingYear
        self.Reversal = Reversal
        self.SalesEntryLines = SalesEntryLines if SalesEntryLines else SalesEntryLineList()
        self.Status = Status
        self.StatusDescription = StatusDescription
        self.Type = Type
        self.TypeDescription = TypeDescription
        self.VATAmountDC = VATAmountDC
        self.VATAmountFC = VATAmountFC
        self.WithholdingTaxAmountDC = WithholdingTaxAmountDC
        self.WithholdingTaxBaseAmount = WithholdingTaxBaseAmount
        self.WithholdingTaxPercentage = WithholdingTaxPercentage
        self.YourRef = YourRef

class SalesEntryLineList(ObjectListModel):
    
    def __init__(self):
        super().__init__(list=[], listObject=SalesEntryLine)

class SalesEntryLine(BaseModel):

    def __init__(self,
        ID=None,
        AmountFC=None,
        Asset=None,
        AssetDescription=None,
        CostCenter=None,
        CostCenterDescription=None,
        CostUnit=None,
        CostUnitDescription=None,
        Description=None,
        Division=None,
        EntryID=None,
        ExtraDutyAmountFC=None,
        ExtraDutyPercentage=None,
        From=None,
        GLAccount=None,
        GLAccountCode=None,
        GLAccountDescription=None,
        IntraStatArea=None,
        IntraStatCountry=None,
        IntraStatDeliveryTerm=None,
        IntraStatTransactionA=None,
        IntraStatTransactionB=None,
        IntraStatTransportMethod=None,
        LineNumber=None,
        Notes=None,
        Project=None,
        ProjectDescription=None,
        Quantity=None,
        SerialNumber=None,
        StatisticalNetWeight=None,
        StatisticalNumber=None,
        StatisticalQuantity=None,
        StatisticalValue=None,
        Subscription=None,
        SubscriptionDescription=None,
        TaxSchedule=None,
        To=None,
        TrackingNumber=None,
        TrackingNumberDescription=None,
        Type=None,
        VATAmountDC=None,
        VATAmountFC=None,
        VATBaseAmountDC=None,
        VATBaseAmountFC=None,
        VATCode=None,
        VATCodeDescription=None,
        VATPercentage=None
    ):

        super().__init__()

        self.ID = ID
        self.AmountFC = AmountFC
        self.Asset = Asset
        self.AssetDescription = AssetDescription
        self.CostCenter = CostCenter
        self.CostCenterDescription = CostCenterDescription
        self.CostUnit = CostUnit
        self.CostUnitDescription = CostUnitDescription
        self.Description = Description
        self.Division = Division
        self.EntryID = EntryID
        self.ExtraDutyAmountFC = ExtraDutyAmountFC
        self.ExtraDutyPercentage = ExtraDutyPercentage
        self.From = From
        self.GLAccount = GLAccount
        self.GLAccountCode = GLAccountCode
        self.GLAccountDescription = GLAccountDescription
        self.IntraStatArea = IntraStatArea
        self.IntraStatCountry = IntraStatCountry
        self.IntraStatDeliveryTerm = IntraStatDeliveryTerm
        self.IntraStatTransactionA = IntraStatTransactionA
        self.IntraStatTransactionB = IntraStatTransactionB
        self.IntraStatTransportMethod = IntraStatTransportMethod
        self.LineNumber = LineNumber
        self.Notes = Notes
        self.Project = Project
        self.ProjectDescription = ProjectDescription
        self.Quantity = Quantity
        self.SerialNumber = SerialNumber
        self.StatisticalNetWeight = StatisticalNetWeight
        self.StatisticalNumber = StatisticalNumber
        self.StatisticalQuantity = StatisticalQuantity
        self.StatisticalValue = StatisticalValue
        self.Subscription = Subscription
        self.SubscriptionDescription = SubscriptionDescription
        self.TaxSchedule = TaxSchedule
        self.To = To
        self.TrackingNumber = TrackingNumber
        self.TrackingNumberDescription = TrackingNumberDescription
        self.Type = Type
        self.VATAmountDC = VATAmountDC
        self.VATAmountFC = VATAmountFC
        self.VATBaseAmountDC = VATBaseAmountDC
        self.VATBaseAmountFC = VATBaseAmountFC
        self.VATCode = VATCode
        self.VATCodeDescription = VATCodeDescription
        self.VATPercentage = VATPercentage


class APIEndpoint:

    def __init__(self, 
        api, 
        endpoint,
        singleObject,
        listObject,
        pkField='ID'
    ):

        self.api = api
        self.endpoint = endpoint
        self.singleObject = singleObject
        self.listObject = listObject
        self.pkField = pkField
    
    def list(self, select=[]):
        url = self.endpoint
        if select: url = '{url}&$select={select}'.format(url=url, select=",".join(select))

        status, headers, respJson = self.api.get(url)
        if status != 200: return self.listObject().parseError(respJson)
        listObj = self.listObject().parse(respJson['d']['results'])

        while('__next' in respJson['d']):
            nextUrl = respJson['d']['__next']
            nextUrl = nextUrl.replace('https://start.exactonline.be/api/v1/{0}/'.format(self.api.division), '')

            status, headers, respJson = self.api.get(nextUrl)
            if status != 200: return self.listObject().parseError(respJson)
            tempListObj = self.listObject().parse(respJson['d']['results'])

            for entryObj in tempListObj.items():
                listObj.add(entryObj)
        
        return listObj

    def get(self, id, select=[]):

        url = "{endpoint}?$filter={pkField} eq guid'{id}'".format(endpoint=self.endpoint, pkField=self.pkField, id=id)
        if select: url = '{url}&$select={select}'.format(url=url, select=",".join(select))

        status, headers, respJson = self.api.get(url)

        if status != 200: return self.singleObject().parseError(respJson)

        return self.singleObject().parse(respJson['d']['results'][0])
    
    def filter(self, field, value, select=[]):
        url = "{endpoint}?$filter={field} eq {value}".format(endpoint=self.endpoint, field=field, value=value)
        if select: url = '{url}&$select={select}'.format(url=url, select=",".join(select))

        status, headers, respJson = self.api.get(url)
        if status != 200: return self.listObject().parseError(respJson)
        listObj = self.listObject().parse(respJson['d']['results'])

        while('__next' in respJson['d']):
            nextUrl = respJson['d']['__next']
            nextUrl = nextUrl.replace('https://start.exactonline.be/api/v1/{0}/'.format(self.api.division), '')

            status, headers, respJson = self.api.get(nextUrl)
            if status != 200: return self.listObject().parseError(respJson)
            tempListObj = self.listObject().parse(respJson['d']['results'])

            for entryObj in tempListObj.items():
                listObj.add(entryObj)
        
        return listObj

    def create(self, object):
        url = self.endpoint
        data = object.getJSON()

        status, headers, respJson = self.api.post(url, data)

        if status not in [200, 201]: return self.singleObject().parseError(respJson)

        return self.singleObject().parse(respJson['d'])
    
    def update(self, object):
        id = getattr(object, self.pkField)
        url = "{endpoint}(guid'{id}')".format(endpoint=self.endpoint, id=id)
        data = object.getJSON()

        status, headers, respJson = self.api.put(url, data)

        if status not in [200, 204]: return self.singleObject().parseError(respJson)
        return True

    def delete(self, id):
        url = "{endpoint}(guid'{id}')".format(endpoint=self.endpoint, id=id)
        data = None

        status, headers, respJson = self.api.delete(url, data)

        if status not in [200, 204]: return self.singleObject().parseError(respJson)
        return True


from .base import ObjectListModel, BaseModel

class DocumentList(ObjectListModel):

    def __init__(self):
        super().__init__(list=[], listObject=Document)

class Document(BaseModel):

    def __init__(self,
        ID=None,
        Account=None,
        AccountCode=None,
        AccountName=None,
        AmountFC=None,
        Body=None,
        Category=None,
        CategoryDescription=None,
        Contact=None,
        ContactFullName=None,
        Created=None,
        Creator=None,
        CreatorFullName=None,
        Currency=None,
        Division=None,
        DocumentDate=None,
        DocumentFolder=None,
        DocumentFolderCode=None,
        DocumentFolderDescription=None,
        DocumentViewUrl=None,
        ExpiryDate=None,
        FinancialTransactionEntryID=None,
        HasEmptyBody=None,
        HID=None,
        Language=None,
        Modified=None,
        Modifier=None,
        ModifierFullName=None,
        Opportunity=None,
        Project=None,
        ProjectCode=None,
        ProjectDescription=None,
        SalesInvoiceNumber=None,
        SalesOrderNumber=None,
        SendMethod=None,
        ShopOrderNumber=None,
        Subject=None,
        Type=None,
        TypeDescription=None,
        Attachments=None
    ):

        super().__init__()

        self.ID = ID
        self.Account = Account
        self.AccountCode = AccountCode
        self.AccountName = AccountName
        self.AmountFC = AmountFC
        self.Body = Body
        self.Category = Category
        self.CategoryDescription = CategoryDescription
        self.Contact = Contact
        self.ContactFullName = ContactFullName
        self.Created = Created
        self.Creator = Creator
        self.CreatorFullName = CreatorFullName
        self.Currency = Currency
        self.Division = Division
        self.DocumentDate = DocumentDate
        self.DocumentFolder = DocumentFolder
        self.DocumentFolderCode = DocumentFolderCode
        self.DocumentFolderDescription = DocumentFolderDescription
        self.DocumentViewUrl = DocumentViewUrl
        self.ExpiryDate = ExpiryDate
        self.FinancialTransactionEntryID = FinancialTransactionEntryID
        self.HasEmptyBody = HasEmptyBody
        self.HID = HID
        self.Language = Language
        self.Modified = Modified
        self.Modifier = Modifier
        self.ModifierFullName = ModifierFullName
        self.Opportunity = Opportunity
        self.Project = Project
        self.ProjectCode = ProjectCode
        self.ProjectDescription = ProjectDescription
        self.SalesInvoiceNumber = SalesInvoiceNumber
        self.SalesOrderNumber = SalesOrderNumber
        self.SendMethod = SendMethod
        self.ShopOrderNumber = ShopOrderNumber
        self.Subject = Subject
        self.Type = Type
        self.TypeDescription = TypeDescription
        self.Attachments = Attachments if Attachments else DocumentAttachmentList()

class DocumentAttachmentList(ObjectListModel):
    def __init__(self):
        super().__init__(list=[], listObject=DocumentAttachment)

class DocumentAttachment(BaseModel):

    def __init__(self,
        ID=None,
        Attachment=None,
        Document=None,
        FileName=None,
        FileSize=None,
        Url=None,
    ):

        self.ID = ID
        self.Attachment = Attachment
        self.Document = Document
        self.FileName = FileName
        self.FileSize = FileSize
        self.Url = Url

class DocumentTypeList(ObjectListModel):
    def __init__(self):
        super().__init__(list=[], listObject=DocumentType)

class DocumentType(BaseModel):

    def __init__(self,
        ID=None,
        Description=None,
        DocumentIsCreatable=None,
        DocumentIsDeletable=None,
        DocumentIsUpdatable=None,
        DocumentIsViewable=None,
        Modified=None,
        TypeCategory=None
    ):


        self.ID = ID
        self.Description = Description
        self.DocumentIsCreatable = DocumentIsCreatable
        self.DocumentIsDeletable = DocumentIsDeletable
        self.DocumentIsUpdatable = DocumentIsUpdatable
        self.DocumentIsViewable = DocumentIsViewable
        self.Modified = Modified
        self.TypeCategory = TypeCategory


class DocumentTypeCategoryList(ObjectListModel):
    def __init__(self):
        super().__init__(list=[], listObject=DocumentTypeCategory)


class DocumentTypeCategory(BaseModel):

    def __init__(self,
        ID=None,
        Created=None,
        Description=None,
        Modified=None
    ):

        self.ID = ID
        self.Created = Created
        self.Description = Description
        self.Modified = Modified


import time
import json

from requests_oauthlib import OAuth2Session

from . import config
from .constants.errors import NotFoundError

class AuthHandler:

    def __init__(self, api, clientId, clientSecret):
        self.clientId = clientId

        self.clientSecret = clientSecret
        self.api = api
        self.cacheHandler = api.cacheHandler

        self.authUrl = config.AUTH_URL
        self.tokenUrl = config.ACCESS_TOKEN_URL
        self.redirectUri = None

        self.state = None
        self.oauthToken = None
        self.token = None
        self.refreshToken = None
        self.expiresOn = 9999999999999

        self.initParams()

    def initParams(self):

        # Check if we already have tokens stored in cache
        # If so, put them in the right variables

        cacheContent = self.cacheHandler.getCache(self.clientId)
        if cacheContent:
            self.token = cacheContent['token']
            self.refreshToken = cacheContent['refreshToken']
            self.expiresOn = cacheContent['expiresOn']

            # Check if token hasnt already expired
            if not self.isTokenDueRenewal():
                self.setTokenHeader(self.token)
            else:
                self.acquireNewToken()

    def isTokenDueRenewal(self):
        currTime = int(time.time())
        if currTime > (self.expiresOn - 30): return True
        return False

    def getAuthURL(self, redirectUri):
        self.redirectUri = redirectUri

        oauth = OAuth2Session(self.clientId, redirect_uri=self.redirectUri)
        authorizationUrl, self.state = oauth.authorization_url(self.authUrl)
        return authorizationUrl

    def tokenSaver(self, oauth):

        self.token = oauth._client.access_token
        self.refreshToken = oauth._client.refresh_token
        self.expiresOn = int(time.time() + float(oauth._client.expires_in))

        cacheContent = { 'token' : self.token, 'refreshToken' : self.refreshToken, 'expiresOn' : self.expiresOn }

        self.cacheHandler.setCache(self.clientId, cacheContent)
        self.setTokenHeader(self.token)

        return self.token
    
    def retrieveToken(self, response, state=None, redirectUri=None):
        if not redirectUri: 
            if not self.redirectUri: raise NotFoundError('redirect uri is not found. init the auth flow first or give the uri as a parameter.')
            redirectUri = self.redirectUri
        
        if not state: 
            if not self.state: raise NotFoundError('state is not found. init the auth flow first or give the state as a parameter.')
            state = self.state

        oauth = OAuth2Session(self.clientId, state=state, redirect_uri=redirectUri)
        self.oauthToken = oauth.fetch_token(self.tokenUrl, client_secret=self.clientSecret, authorization_response=response)

        token = self.tokenSaver(oauth)

        return token
    
    def acquireNewToken(self):

        tempToken = {
            'access_token' : self.token,
            'refresh_token' : self.refreshToken,
            'token_type' : 'Bearer',
            'expires_in' : '-30'
        }

        oauth = OAuth2Session(self.clientId, token=tempToken)
        
        # Try to get a new token with the cached access & refresh tokens
        # If refresh token also expired, remove cache so flow can be init again
        
        try:
            self.oauthToken = oauth.refresh_token(self.tokenUrl, client_id=self.clientId, client_secret=self.clientSecret)
        except:
            self.cacheHandler.deleteCache(self.clientId)
            raise NotFoundError('refresh token is not found. this is likely because the refresh token has expired as well. init the auth flow again.')

        token = self.tokenSaver(oauth)

        return token
    
    def getToken(self):

        if self.token: return self.token
        raise NotFoundError('token is not found. init the auth flow first.')
    
    def setTokenHeader(self, token):
        bearerStr = 'Bearer {token}'.format(token=token)
        self.api.headers.update({'Authorization' : bearerStr})
    
    def checkHeaderTokens(self):

        # Check if token is due for renewal
        if self.isTokenDueRenewal():
            token = self.acquireNewToken()

        # If no authorization header is found, we need to include the token
        if 'Authorization' not in self.api.headers:
            
            # Check if we have a token stored in cache, if not, acquire one
            # If we do, set it in the header
            cacheContent = self.cacheHandler.getCache(self.clientId)
            token = cacheContent['token'] if cacheContent else None

            if token is None: token = self.getToken()
            self.setTokenHeader(token)


from .base import APIEndpoint

from exactonline.models.glaccounts import GLAccountList, GLAccount

class GLAccountMethods(APIEndpoint):

    def __init__(self, api):
        super().__init__(api, 'financial/GLAccounts', GLAccount, GLAccountList)


from .base import APIEndpoint

from exactonline.models.journals import Journal, JournalList

class JournalMethods(APIEndpoint):

    def __init__(self, api):
        super().__init__(api, 'financial/Journals', Journal, JournalList)


from .base import APIEndpoint

from exactonline.models.vat import VATPercentageList, VATPercentage, VATCode, VATCodeList

class VATCodeMethods(APIEndpoint):

    def __init__(self, api):
        super().__init__(api, 'vat/VATCodes', VATCode, VATCodeList)
    

    def get(self, id, select=[]):

        # 1: retrieve VATCode
        vatCode = super().get(id, select)

        # 2: retrieve Percentages
        url = "{endpoint}(guid'{entryId}')/VATPercentages".format(endpoint=self.endpoint, entryId=vatCode.ID)
        status, headers, respJson = self.api.get(url)

        if status != 200: return VATPercentageList().parseError(respJson)
        VATPercentages = VATPercentageList().parse(respJson['d']['results'])

        vatCode.VATPercentages = VATPercentages

        return vatCode


import base64
import requests
import json
import time

from . import config
from .cachehandler import CacheHandler
from .authhandler import AuthHandler

from .endpoints.salesentries import SalesEntryMethods
from .endpoints.accounts import AccountMethods
from .endpoints.journals import JournalMethods
from .endpoints.glaccounts import GLAccountMethods
from .endpoints.documents import DocumentMethods
from .endpoints.contacts import ContactMethods
from .endpoints.vatcodes import VATCodeMethods

class ExactOnlineAPI:

    def __init__(self, clientId, clientSecret):

        self.clientId = clientId
        self.clientSecret = clientSecret
        self.division = None

        self.headers = {
            'Accept' : 'application/json',
            'Content-Type' : 'application/json',
        }

        self.baseUrl = config.BASE_URL
        self.cacheHandler = CacheHandler()
        self.authHandler = AuthHandler(self, self.clientId, self.clientSecret)

        self.salesEntries = SalesEntryMethods(self)
        self.accounts = AccountMethods(self)
        self.journals = JournalMethods(self)
        self.glAccounts = GLAccountMethods(self)
        self.documents = DocumentMethods(self)
        self.contacts = ContactMethods(self)
        self.vatCodes = VATCodeMethods(self)

    def doRequest(self, method, url, data=None, headers=None, files=None):

        if headers:
            mergedHeaders = self.headers
            mergedHeaders.update(headers)
            headers = mergedHeaders
        else: headers = self.headers

        reqUrl = '{base}/{division}/{url}'.format(base=self.baseUrl, division=self.division, url=url)

        if method == 'GET':
            response = requests.get(reqUrl, params=data, headers=headers)
        elif method == 'POST':
            if files: response = requests.post(reqUrl, data=json.dumps(data), files=files, headers=headers)
            else: response = requests.post(reqUrl, data=json.dumps(data), headers=headers)
        elif method == 'PUT':
            response = requests.put(reqUrl, data=json.dumps(data), headers=headers)
        elif method == 'DELETE':
            response = requests.delete(reqUrl, params=json.dumps(data), headers=headers)

        return response

    def request(self, method, url, data=None, headers=None, files=None):

        self.authHandler.checkHeaderTokens()
        self.checkDivision()

        response = self.doRequest(method, url, data, headers, files)
        respContent = json.loads(response.content) if response.content else None

        return response.status_code, response.headers, respContent
    
    def checkDivision(self):
        if not self.division:
            url = '{base}/current/Me?$select=CurrentDivision'.format(base=self.baseUrl)
            response = requests.get(url, params=None, headers=self.headers)

            jsonResp = json.loads(response.content)
            self.division = jsonResp['d']['results'][0]['CurrentDivision']

    def get(self, url, data=None, headers=None):
        status, headers, response = self.request('GET', url, data, headers)
        return status, headers, response
    
    def post(self, url, data=None, headers=None, files=None):
        status, headers, response = self.request('POST', url, data, headers, files)
        return status, headers, response
    
    def put(self, url, data=None, headers=None):
        status, headers, response = self.request('PUT', url, data, headers)
        return status, headers, response
    
    def delete(self, url, data=None, headers=None):
        status, headers, response = self.request('DELETE', url, data, headers)
        return status, headers, response


import json
import os

from pathlib import Path

from . import config

class CacheHandler:

    def __init__(self):
        self.rootDir = Path(__file__).parent.absolute()
        self.cacheDir = '{root}/cache'.format(root=self.rootDir)

    def resetCache(self):
        preserveFile = '__init__.py'
        for f in sorted(os.listdir(self.cacheDir)):
            if f != preserveFile:
                fullPath = os.path.join(self.cacheDir, f)
                os.remove(fullPath)
        
        config.CACHE = {}

        return config.CACHE

    def getCache(self, key):
        # if key in config.CACHE: return config.CACHE[key]

        keyPath = '{cache}/{key}.txt'.format(cache=self.cacheDir, key=key)

        try:
            with open(keyPath, 'r') as f:
                value = json.load(f)

            config.CACHE[key] = value
            return value

        except IOError:
            return None
    
    def setCache(self, key, value):
        keyPath = '{cache}/{key}.txt'.format(cache=self.cacheDir, key=key)
        value = json.dumps(value)

        try:
            with open(keyPath, 'w') as f:
                f.write(value)
            
            config.CACHE[key] = value
            return value

        except IOError as e:
            raise IOError('couldnt save to cache. error raised: {0}'.format(e))
    
    def deleteCache(self, key):
        keyPath = '{cache}/{key}.txt'.format(cache=self.cacheDir, key=key)
        
        try:
            if os.path.exists(keyPath):
                os.remove(keyPath)
        except IOError as e:
            raise IOError('couldnt delete cache. error raised: {0}'.format(e)) 


from .base import APIEndpoint

from exactonline.models.salesentries import SalesEntryLineList, SalesEntryList, SalesEntry, SalesEntryLine

class SalesEntryMethods(APIEndpoint):

    def __init__(self, api):
        super().__init__(api, 'salesentry/SalesEntries', SalesEntry, SalesEntryList, pkField='EntryID')

    def get(self, id, select=[]):

        # 1: retrieve Sales Entry
        salesEntry = super().get(id, select)

        # 2: retrieve Sales Lines
        url = "{endpoint}(guid'{entryId}')/SalesEntryLines".format(endpoint=self.endpoint, entryId=salesEntry.EntryID)
        status, headers, respJson = self.api.get(url)

        if status != 200: return SalesEntryLineList().parseError(respJson)
        salesEntryLines = SalesEntryLineList().parse(respJson['d']['results'])

        salesEntry.SalesEntryLines = salesEntryLines

        return salesEntry


from .base import ObjectListModel, BaseModel

class ContactList(ObjectListModel):

    def __init__(self):
        super().__init__(list=[], listObject=Contact)

class Contact(BaseModel):

    def __init__(self,
        ID=None,
        Account=None,
        AccountIsCustomer=None,
        AccountIsSupplier=None,
        AccountMainContact=None,
        AccountName=None,
        AddressLine2=None,
        AddressStreet=None,
        AddressStreetNumber=None,
        AddressStreetNumberSuffix=None,
        AllowMailing=None,
        BirthDate=None,
        BirthName=None,
        BirthNamePrefix=None,
        BirthPlace=None,
        BusinessEmail=None,
        BusinessFax=None,
        BusinessMobile=None,
        BusinessPhone=None,
        BusinessPhoneExtension=None,
        City=None,
        Code=None,
        Country=None,
        Created=None,
        Creator=None,
        CreatorFullName=None,
        Division=None,
        Email=None,
        EndDate=None,
        FirstName=None,
        FullName=None,
        Gender=None,
        HID=None,
        IdentificationDate=None,
        IdentificationDocument=None,
        IdentificationUser=None,
        Initials=None,
        IsAnonymised=None,
        IsMailingExcluded=None,
        IsMainContact=None,
        JobTitleDescription=None,
        Language=None,
        LastName=None,
        LeadPurpose=None,
        LeadSource=None,
        MarketingNotes=None,
        MiddleName=None,
        Mobile=None,
        Modified=None,
        Modifier=None,
        ModifierFullName=None,
        Nationality=None,
        Notes=None,
        PartnerName=None,
        PartnerNamePrefix=None,
        Person=None,
        Phone=None,
        PhoneExtension=None,
        Picture=None,
        PictureName=None,
        PictureThumbnailUrl=None,
        PictureUrl=None,
        Postcode=None,
        SocialSecurityNumber=None,
        StartDate=None,
        State=None,
        Title=None
    ):

        super().__init__()
        
        self.ID = ID
        self.Account = Account
        self.AccountIsCustomer = AccountIsCustomer
        self.AccountIsSupplier = AccountIsSupplier
        self.AccountMainContact = AccountMainContact
        self.AccountName = AccountName
        self.AddressLine2 = AddressLine2
        self.AddressStreet = AddressStreet
        self.AddressStreetNumber = AddressStreetNumber
        self.AddressStreetNumberSuffix = AddressStreetNumberSuffix
        self.AllowMailing = AllowMailing
        self.BirthDate = BirthDate
        self.BirthName = BirthName
        self.BirthNamePrefix = BirthNamePrefix
        self.BirthPlace = BirthPlace
        self.BusinessEmail = BusinessEmail
        self.BusinessFax = BusinessFax
        self.BusinessMobile = BusinessMobile
        self.BusinessPhone = BusinessPhone
        self.BusinessPhoneExtension = BusinessPhoneExtension
        self.City = City
        self.Code = Code
        self.Country = Country
        self.Created = Created
        self.Creator = Creator
        self.CreatorFullName = CreatorFullName
        self.Division = Division
        self.Email = Email
        self.EndDate = EndDate
        self.FirstName = FirstName
        self.FullName = FullName
        self.Gender = Gender
        self.HID = HID
        self.IdentificationDate = IdentificationDate
        self.IdentificationDocument = IdentificationDocument
        self.IdentificationUser = IdentificationUser
        self.Initials = Initials
        self.IsAnonymised = IsAnonymised
        self.IsMailingExcluded = IsMailingExcluded
        self.IsMainContact = IsMainContact
        self.JobTitleDescription = JobTitleDescription
        self.Language = Language
        self.LastName = LastName
        self.LeadPurpose = LeadPurpose
        self.LeadSource = LeadSource
        self.MarketingNotes = MarketingNotes
        self.MiddleName = MiddleName
        self.Mobile = Mobile
        self.Modified = Modified
        self.Modifier = Modifier
        self.ModifierFullName = ModifierFullName
        self.Nationality = Nationality
        self.Notes = Notes
        self.PartnerName = PartnerName
        self.PartnerNamePrefix = PartnerNamePrefix
        self.Person = Person
        self.Phone = Phone
        self.PhoneExtension = PhoneExtension
        self.Picture = Picture
        self.PictureName = PictureName
        self.PictureThumbnailUrl = PictureThumbnailUrl
        self.PictureUrl = PictureUrl
        self.Postcode = Postcode
        self.SocialSecurityNumber = SocialSecurityNumber
        self.StartDate = StartDate
        self.State = State
        self.Title = Title


from setuptools import setup

# read the contents of README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
  name = 'python-exact-online',         
  packages=['exactonline', 'exactonline.models', 'exactonline.constants', 'exactonline.cache', 'exactonline.endpoints'],
  version = '0.3.3',
  license='GPL-3.0-or-later',
  description = 'Basic wrapper for the Exact Online REST API (v1)',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'Alexander Schillemans',
  author_email = 'alexander.schillemans@lhs.global',
  url = 'https://github.com/alexanderlhsglobal/python-exact-online',
  download_url = 'https://github.com/alexanderlhsglobal/python-exact-online/archive/refs/tags/0.3.3.tar.gz',
  keywords = ['exact', 'exact online', 'api'],
  install_requires=[
          'requests',
          'oauthlib',
          'requests_oauthlib',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    'Programming Language :: Python :: 3.6',
  ],
)


from .base import ObjectListModel, BaseModel

class AccountList(ObjectListModel):

    def __init__(self):
        super().__init__(list=[], listObject=Account)

class Account(BaseModel):

    def __init__(self,
        ID=None,
        Accountant=None,
        AccountManager=None,
        AccountManagerFullName=None,
        AccountManagerHID=None,
        ActivitySector=None,
        ActivitySubSector=None,
        AddressLine1=None,
        AddressLine2=None,
        AddressLine3=None,
        BankAccounts=None,
        Blocked=None,
        BSN=None,
        BusinessType=None,
        CanDropShip=None,
        ChamberOfCommerce=None,
        City=None,
        Classification1=None,
        Classification2=None,
        Classification3=None,
        Classification4=None,
        Classification5=None,
        Classification6=None,
        Classification7=None,
        Classification8=None,
        Code=None,
        CodeAtSupplier=None,
        CompanySize=None,
        ConsolidationScenario=None,
        ControlledDate=None,
        Country=None,
        CountryName=None,
        Created=None,
        Creator=None,
        CreatorFullName=None,
        CreditLinePurchase=None,
        CreditLineSales=None,
        DatevCreditorCode=None,
        DatevDebtorCode=None,
        DiscountPurchase=None,
        DiscountSales=None,
        Division=None,
        Email=None,
        EndDate=None,
        EstablishedDate=None,
        Fax=None,
        GLAccountPurchase=None,
        GLAccountSales=None,
        GLAP=None,
        GLAR=None,
        GlnNumber=None,
        HasWithholdingTaxSales=None,
        ignoreDatevWarningMessage=None,
        IntraStatArea=None,
        IntraStatDeliveryTerm=None,
        IntraStatTransactionA=None,
        IntraStatTractionB=None,
        IntraStatTransporterMethod=None,
        InvoiceAccount=None,
        InvoiceAccountCode=None,
        InvoiceAccountName=None,
        InvoiceAttachmentType=None,
        InvoicingMethod=None,
        IsAccountant=None,
        IsAgency=None,
        IsAnonymised=None,
        IsCompetitor=None,
        IsExtraDuty=None,
        IsMailing=None,
        IsPilot=None,
        IsReseller=None,
        IsSales=None,
        IsSupplier=None,
        Language=None,
        LanguageDescription=None,
        Latitude=None,
        LeadPurpose=None,
        LeadSource=None,
        Logo=None,
        LogoFileName=None,
        LogoThumbnailUrl=None,
        LogoUrl=None,
        Longitude=None,
        MainContact=None,
        Modified=None,
        Modifier=None,
        ModifierFullName=None,
        Name=None,
        OINNumber=None,
        Parent=None,
        PayAsYouEarn=None,
        PaymentConditionPurchase=None,
        PaymentConditionPurchaseDescription=None,
        PaymentConditionSales=None,
        PaymentConditionSalesDescription=None,
        Phone=None,
        PhoneExtension=None,
        Postcode=None,
        PriceList=None,
        PurchaseCurrency=None,
        PurchaseCurrencyDescription=None,
        PurchaseLeadDays=None,
        PurchaseVATCode=None,
        PurchaseVATCodeDescription=None,
        RecepientOfCommissions=None,
        Remarks=None,
        Reseller=None,
        ResellerCode=None,
        ResellerName=None,
        RSIN=None,
        SalesCurrency=None,
        SalesCurrencyDescription=None,
        SalesVATCode=None,
        SalesVATCodeDescription=None,
        SearchCode=None,
        SecurityLevel=None,
        SeperateInvPerSubscription=None,
        ShippingLeadDays=None,
        ShippingMethod=None,
        ShowRemarkForSales=None,
        StartDate=None,
        State=None,
        StateName=None,
        Status=None,
        StatusSince=None,
        TradeName=None,
        Type=None,
        UniqueTaxpayerReference=None,
        VATLiability=None,
        VATNumber=None,
        Website=None
    ):

        super().__init__()

        self.ID = ID
        self.Accountant = Accountant
        self.AccountManager = AccountManager
        self.AccountManagerFullName = AccountManagerFullName
        self.AccountManagerHID = AccountManagerHID
        self.ActivitySector = ActivitySector
        self.ActivitySubSector = ActivitySubSector
        self.AddressLine1 = AddressLine1
        self.AddressLine2 = AddressLine2
        self.AddressLine3 = AddressLine3
        self.BankAccounts = BankAccounts
        self.Blocked = Blocked
        self.BSN = BSN
        self.BusinessType = BusinessType
        self.CanDropShip = CanDropShip
        self.ChamberOfCommerce = ChamberOfCommerce
        self.City = City
        self.Classification1 = Classification1
        self.Classification2 = Classification2
        self.Classification3 = Classification3
        self.Classification4 = Classification4
        self.Classification5 = Classification5
        self.Classification6 = Classification6
        self.Classification7 = Classification7
        self.Classification8 = Classification8
        self.Code = Code
        self.CodeAtSupplier = CodeAtSupplier
        self.CompanySize = CompanySize
        self.ConsolidationScenario = ConsolidationScenario
        self.ControlledDate = ControlledDate
        self.Country = Country
        self.CountryName = CountryName
        self.Created = Created
        self.Creator = Creator
        self.CreatorFullName = CreatorFullName
        self.CreditLinePurchase = CreditLinePurchase
        self.CreditLineSales = CreditLineSales
        self.DatevCreditorCode = DatevCreditorCode
        self.DatevDebtorCode = DatevDebtorCode
        self.DiscountPurchase = DiscountPurchase
        self.DiscountSales = DiscountSales
        self.Division = Division
        self.Email = Email
        self.EndDate = EndDate
        self.EstablishedDate = EstablishedDate
        self.Fax = Fax
        self.GLAccountPurchase = GLAccountPurchase
        self.GLAccountSales = GLAccountSales
        self.GLAP = GLAP
        self.GLAR = GLAR
        self.GlnNumber = GlnNumber
        self.HasWithholdingTaxSales = HasWithholdingTaxSales
        self.ignoreDatevWarningMessage = ignoreDatevWarningMessage
        self.IntraStatArea = IntraStatArea
        self.IntraStatDeliveryTerm = IntraStatDeliveryTerm
        self.IntraStatTransactionA = IntraStatTransactionA
        self.IntraStatTractionB = IntraStatTractionB
        self.IntraStatTransporterMethod = IntraStatTransporterMethod
        self.InvoiceAccount = InvoiceAccount
        self.InvoiceAccountCode = InvoiceAccountCode
        self.InvoiceAccountName = InvoiceAccountName
        self.InvoiceAttachmentType = InvoiceAttachmentType
        self.InvoicingMethod = InvoicingMethod
        self.IsAccountant = IsAccountant
        self.IsAgency = IsAgency
        self.IsAnonymised = IsAnonymised
        self.IsCompetitor = IsCompetitor
        self.IsExtraDuty = IsExtraDuty
        self.IsMailing = IsMailing
        self.IsPilot = IsPilot
        self.IsReseller = IsReseller
        self.IsSales = IsSales
        self.IsSupplier = IsSupplier
        self.Language = Language
        self.LanguageDescription = LanguageDescription
        self.Latitude = Latitude
        self.LeadPurpose = LeadPurpose
        self.LeadSource = LeadSource
        self.Logo = Logo
        self.LogoFileName = LogoFileName
        self.LogoThumbnailUrl = LogoThumbnailUrl
        self.LogoUrl = LogoUrl
        self.Longitude = Longitude
        self.MainContact = MainContact
        self.Modified = Modified
        self.Modifier = Modifier
        self.ModifierFullName = ModifierFullName
        self.Name = Name
        self.OINNumber = OINNumber
        self.Parent = Parent
        self.PayAsYouEarn = PayAsYouEarn
        self.PaymentConditionPurchase = PaymentConditionPurchase
        self.PaymentConditionPurchaseDescription = PaymentConditionPurchaseDescription
        self.PaymentConditionSales = PaymentConditionSales
        self.PaymentConditionSalesDescription = PaymentConditionSalesDescription
        self.Phone = Phone
        self.PhoneExtension = PhoneExtension
        self.Postcode = Postcode
        self.PriceList = PriceList
        self.PurchaseCurrency = PurchaseCurrency
        self.PurchaseCurrencyDescription = PurchaseCurrencyDescription
        self.PurchaseLeadDays = PurchaseLeadDays
        self.PurchaseVATCode = PurchaseVATCode
        self.PurchaseVATCodeDescription = PurchaseVATCodeDescription
        self.RecepientOfCommissions = RecepientOfCommissions
        self.Remarks = Remarks
        self.Reseller = Reseller
        self.ResellerCode = ResellerCode
        self.ResellerName = ResellerName
        self.RSIN = RSIN
        self.SalesCurrency = SalesCurrency
        self.SalesCurrencyDescription = SalesCurrencyDescription
        self.SalesVATCode = SalesVATCode
        self.SalesVATCodeDescription = SalesVATCodeDescription
        self.SearchCode = SearchCode
        self.SecurityLevel = SecurityLevel
        self.SeperateInvPerSubscription = SeperateInvPerSubscription
        self.ShippingLeadDays = ShippingLeadDays
        self.ShippingMethod = ShippingMethod
        self.ShowRemarkForSales = ShowRemarkForSales
        self.StartDate = StartDate
        self.State = State
        self.StateName = StateName
        self.Status = Status
        self.StatusSince = StatusSince
        self.TradeName = TradeName
        self.Type = Type
        self.UniqueTaxpayerReference = UniqueTaxpayerReference
        self.VATLiability = VATLiability
        self.VATNumber = VATNumber
        self.Website = Website


from .base import ObjectListModel, BaseModel

class GLAccountList(ObjectListModel):

    def __init__(self):
        super().__init__(list=[], listObject=GLAccount)

class GLAccount(BaseModel):

    def __init__(self,
        ID=None,
        AssimilatedVATBox=None,
        BalanceSide=None,
        BalanceType=None,
        BelcotaxType=None,
        Code=None,
        Compress=None,
        Costcenter=None,
        CostcenterDescription=None,
        Costunit=None,
        CostunitDescription=None,
        Created=None,
        Creator=None,
        CreatorFullName=None,
        DeductibilityPercentages=None,
        Description=None,
        Division=None,
        ExcludeVATListing=None,
        ExpenseNonDeductiblePercentage=None,
        IsBlocked=None,
        Matching=None,
        Modified=None,
        Modifier=None,
        ModifierFullName=None,
        PrivateGLAcount=None,
        PrivatePercentage=None,
        ReportingCode=None,
        RevalueCurrency=None,
        SearchCode=None,
        Type=None,
        TypeDescription=None,
        UseCostcenter=None,
        UseCostunit=None,
        VATCode=None,
        VATDescription=None,
        VATGLAccountType=None,
        VATNonDeductibleGLAccount=None,
        VATNonDeductiblePercentage=None,
        VATSystem=None,
        YearEndCostGLAccount=None,
        YearEndReflectionGLAccount=None
    ):

        super().__init__()

        self.ID = ID
        self.AssimilatedVATBox = AssimilatedVATBox
        self.BalanceSide = BalanceSide
        self.BalanceType = BalanceType
        self.BelcotaxType = BelcotaxType
        self.Code = Code
        self.Compress = Compress
        self.Costcenter = Costcenter
        self.CostcenterDescription = CostcenterDescription
        self.Costunit = Costunit
        self.CostunitDescription = CostunitDescription
        self.Created = Created
        self.Creator = Creator
        self.CreatorFullName = CreatorFullName
        self.DeductibilityPercentages = DeductibilityPercentages
        self.Description = Description
        self.Division = Division
        self.ExcludeVATListing = ExcludeVATListing
        self.ExpenseNonDeductiblePercentage = ExpenseNonDeductiblePercentage
        self.IsBlocked = IsBlocked
        self.Matching = Matching
        self.Modified = Modified
        self.Modifier = Modifier
        self.ModifierFullName = ModifierFullName
        self.PrivateGLAcount = PrivateGLAcount
        self.PrivatePercentage = PrivatePercentage
        self.ReportingCode = ReportingCode
        self.RevalueCurrency = RevalueCurrency
        self.SearchCode = SearchCode
        self.Type = Type
        self.TypeDescription = TypeDescription
        self.UseCostcenter = UseCostcenter
        self.UseCostunit = UseCostunit
        self.VATCode = VATCode
        self.VATDescription = VATDescription
        self.VATGLAccountType = VATGLAccountType
        self.VATNonDeductibleGLAccount = VATNonDeductibleGLAccount
        self.VATNonDeductiblePercentage = VATNonDeductiblePercentage
        self.VATSystem = VATSystem
        self.YearEndCostGLAccount = YearEndCostGLAccount
        self.YearEndReflectionGLAccount = YearEndReflectionGLAccount


from .base import ObjectListModel, BaseModel

class VATPercentageList(ObjectListModel):

    def __init__(self):
        super().__init__(list=[], listObject=VATPercentage)
    

class VATPercentage(BaseModel):

    def __init__(self,
        ID=None,
        VATCodeID=None,
        Type=None,
        StartDate=None,
        Percentage=None,
        ModifierFullName=None,
        Modifier=None,
        Modified=None,
        LineNumber=None,
        EndDate=None,
        Division=None,
        CreatorFullName=None,
        Creator=None,
        Created=None
    ):

        super().__init__()

        self.ID = ID
        self.VATCodeID = VATCodeID
        self.Type = Type
        self.StartDate = StartDate
        self.Percentage = Percentage
        self.ModifierFullName = ModifierFullName
        self.Modifier = Modifier
        self.Modified = Modified
        self.LineNumber = LineNumber
        self.EndDate = EndDate
        self.Division = Division
        self.CreatorFullName = CreatorFullName
        self.Creator = Creator
        self.Created = Created

class VATCodeList(ObjectListModel):

    def __init__(self):
        super().__init__(list=[], listObject=VATCode)

class VATCode(BaseModel):

    def __init__(self,
        ID=None,
        Account=None,
        AccountCode=None,
        AccountName=None,
        CalculationBasis=None,
        Charged=None,
        Code=None,
        Country=None,
        Created=None,
        Creator=None,
        CreatorFullName=None,
        Description=None,
        Division=None,
        EUSalesListing=None,
        GLDiscountPurchase=None,
        GLDiscountPurchaseCode=None,
        GLDiscountPurchaseDescription=None,
        GLDiscountSales=None,
        GLDiscountSalesCode=None,
        GLDiscountSalesDescription=None,
        GLToClaim=None,
        GLToClaimCode=None,
        GLToClaimDescription=None,
        GLToPay=None,
        GLToPayCode=None,
        GLToPayDescription=None,
        IntraStat=None,
        IntrastatType=None,
        IsBlocked=None,
        LegalText=None,
        Modified=None,
        Modifier=None,
        ModifierFullName=None,
        OssCountry=None,
        Percentage=None,
        TaxReturnType=None,
        Type=None,
        VatDocType=None,
        VatMargin=None,
        VATPartialRatio=None,
        VATPercentages=None,
        VATTransactionType=None
    ):

        super().__init__()

        self.ID = ID
        self.Account = Account
        self.AccountCode = AccountCode
        self.AccountName = AccountName
        self.CalculationBasis = CalculationBasis
        self.Charged = Charged
        self.Code = Code
        self.Country = Country
        self.Created = Created
        self.Creator = Creator
        self.CreatorFullName = CreatorFullName
        self.Description = Description
        self.Division = Division
        self.EUSalesListing = EUSalesListing
        self.GLDiscountPurchase = GLDiscountPurchase
        self.GLDiscountPurchaseCode = GLDiscountPurchaseCode
        self.GLDiscountPurchaseDescription = GLDiscountPurchaseDescription
        self.GLDiscountSales = GLDiscountSales
        self.GLDiscountSalesCode = GLDiscountSalesCode
        self.GLDiscountSalesDescription = GLDiscountSalesDescription
        self.GLToClaim = GLToClaim
        self.GLToClaimCode = GLToClaimCode
        self.GLToClaimDescription = GLToClaimDescription
        self.GLToPay = GLToPay
        self.GLToPayCode = GLToPayCode
        self.GLToPayDescription = GLToPayDescription
        self.IntraStat = IntraStat
        self.IntrastatType = IntrastatType
        self.IsBlocked = IsBlocked
        self.LegalText = LegalText
        self.Modified = Modified
        self.Modifier = Modifier
        self.ModifierFullName = ModifierFullName
        self.OssCountry = OssCountry
        self.Percentage = Percentage
        self.TaxReturnType = TaxReturnType
        self.Type = Type
        self.VatDocType = VatDocType
        self.VatMargin = VatMargin
        self.VATPartialRatio = VATPartialRatio
        self.VATPercentages = VATPercentages if VATPercentages else VATPercentageList()
        self.VATTransactionType = VATTransactionType